/***********************************************************************************************
GbxLightMap - Extracts the Lightmaps from a given challenge file
Copyright (c) 2013-2018 by Electron

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

This program includes code from the LZO real-time data compression
library. This portion of code is also subject to its own copyright
and licensing conditions, which can be found in the miniLZO directory.

Revision History:
2013-06-02 : Electron : Original version.
2017-03-10 : Electron : Added support for chunk version 5.
2017-07-05 : Electron : Added support for multiple lightmaps.
2018-10-27 : Electron : Added support for chunk version 6.

***********************************************************************************************/

#ifndef STRICT
#define STRICT 1
#endif
#define WIN32_LEAN_AND_MEAN
#define _CRT_SECURE_NO_DEPRECATE

#include <windows.h>
#include <windowsx.h>
#include <shellapi.h>
#include <shlwapi.h>
#include <tchar.h>

#include "minilzo\minilzo.h"

/**********************************************************************************************/

/* Magic 'GBX' */
#define GBX_MAGIC 0x584247

/* Challenge Class ID's */
#define CLSID_CHALLENGE_TMF  0x03043000
#define CLSID_CHALLENGE_TM   0x24003000
#define CLSID_CHALLENGE_VSK  0x21080000

/**********************************************************************************************/

const TCHAR g_szGbx[]        = TEXT(".Gbx");
const TCHAR g_szMapGbx[]     = TEXT(".Map.Gbx");
const TCHAR g_szWebp[]       = TEXT(".webp");
const TCHAR g_szJpeg[]       = TEXT(".jpg");
const TCHAR g_szFormat[]     = TEXT(".LightMap%u");
const TCHAR g_szTitle[]      = TEXT("Lightmap Extractor");
const TCHAR g_szUsage[]      = TEXT("Usage: GbxLightMap.exe <Map.Gbx>");
const TCHAR g_szErrOom[]     = TEXT("Out of memory.");
const TCHAR g_szErrSrc[]     = TEXT("Source file not found.");
const TCHAR g_szErrOpen[]    = TEXT("Could not open source file.");
const TCHAR g_szErrCreate[]  = TEXT("Could not create destination file.");
const TCHAR g_szErrRead[]    = TEXT("Error while reading source file.");
const TCHAR g_szErrWrite[]   = TEXT("Error while writing destination file.");
const TCHAR g_szErrFormat[]  = TEXT("Invalid or unsupported file format.");
const TCHAR g_szErrUncomp[]  = TEXT("Uncompressed maps not supported.");
const TCHAR g_szErrDecomp[]  = TEXT("Could not decompress data section.");
const TCHAR g_szErrChunk[]   = TEXT("Lightmap chunk not found.");
const TCHAR g_szErrVersion[] = TEXT("Unsupported lightmap chunk version.");
const TCHAR g_szErrNone[]    = TEXT("No lightmaps found.");

#define MsgBox(Text) MessageBox(NULL, Text, g_szTitle, MB_OK | MB_ICONEXCLAMATION)

BOOL WriteLightMapFile(PTSTR pszFilename, PBYTE* pData, DWORD* dwCount);

/**********************************************************************************************/

int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	TCHAR szFileGbx[MAX_PATH];
	TCHAR szFileName[MAX_PATH];
	HANDLE hFileInput = NULL;
	PBYTE pDataFile = NULL;
	PBYTE pDataBlock = NULL;
	PBYTE pData = NULL;
	DWORD dwData = 0;
	DWORD dwVersion = 0;
	DWORD dwCount = 0;
	DWORD dwFrames = 1;
	DWORD dwFileSize = 0;
	DWORD dwHeaderSize = 0;
	DWORD dwBytesRead = 0;
	DWORD dwDataFileSize = 0;
	DWORD dwDataBlockSize = 0;
	BOOL  bFound = FALSE;
	PTSTR pch = NULL;

	/* The source file must be passed as program argument */
	if (lpCmdLine == NULL || lpCmdLine[0] == '\0')
	{
		MsgBox(g_szUsage);
		return 0;
	}

	_tcsncpy(szFileGbx, (LPCTSTR)lpCmdLine, MAX_PATH-1);

	/* Remove quotation marks */
	PathUnquoteSpaces(szFileGbx);

	/* Does the passed file exist? */
	if (!PathFileExists(szFileGbx))
	{
		MsgBox(g_szErrSrc);
		return 0;
	}

	/* Build the common part of the output path */
	_tcsncpy(szFileName, szFileGbx, MAX_PATH-1);
	if (pch = StrStrI(szFileName, g_szMapGbx))
		*pch = TEXT('\0');
	else if (pch = StrStrI(szFileName, g_szGbx))
		*pch = TEXT('\0');

	/* Open source file */
	hFileInput = CreateFile(szFileGbx, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFileInput == INVALID_HANDLE_VALUE)
	{
		MsgBox(g_szErrOpen);
		return 0;
	}

	/* Retrieve file size */
	dwFileSize = GetFileSize(hFileInput, NULL);

	/* Read and check magic 'GBX' */
	if (!ReadFile(hFileInput, (LPVOID)&dwData, 3, &dwBytesRead, NULL) ||
		dwBytesRead != 3 || (dwData & 0x00FFFFFF) != GBX_MAGIC)
	{
		MsgBox(g_szErrFormat);
		goto Exit;
	}

	/* Jump to storage settings */
	if (SetFilePointer(hFileInput, 2, NULL, FILE_CURRENT) == 0xFFFFFFFF)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read storage settings */
	if (!ReadFile(hFileInput, (LPVOID)&dwData, 4, &dwBytesRead, NULL) || dwBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Is this an uncompressed gamebox file? */
	if ((dwData & 0x00FF0000) == 0x00550000)
	{
		MsgBox(g_szErrUncomp);
		goto Exit;
	}

	/* Read Class ID */
	if (!ReadFile(hFileInput, (LPVOID)&dwData, 4, &dwBytesRead, NULL) || dwBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Is this a challenge? */
	if (dwData != CLSID_CHALLENGE_TMF && dwData != CLSID_CHALLENGE_TM && dwData != CLSID_CHALLENGE_VSK)
	{
		MsgBox(g_szErrFormat);
		goto Exit;
	}

	/* Read header size */
	if (!ReadFile(hFileInput, (LPVOID)&dwHeaderSize, 4, &dwBytesRead, NULL) || dwBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Jump directly to the block size information in the data section */
	if (SetFilePointer(hFileInput, dwHeaderSize + 8, NULL, FILE_CURRENT) == 0xFFFFFFFF)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read size of uncompressed data section */
	if (!ReadFile(hFileInput, (LPVOID)&dwDataBlockSize, 4, &dwBytesRead, NULL) || dwBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read size of compressed data section */
	if (!ReadFile(hFileInput, (LPVOID)&dwDataFileSize, 4, &dwBytesRead, NULL) || dwBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Sanity check */
	if (dwDataFileSize > dwFileSize)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Allocate memory for the compressed data */
	pDataFile = GlobalAllocPtr(GHND, dwDataFileSize);
	if (pDataFile == NULL)
	{
		MsgBox(g_szErrOom);
		goto Exit;
	}

	/* Allocate memory for the uncompressed data */
	pDataBlock = GlobalAllocPtr(GHND, dwDataBlockSize);
	if (pDataBlock == NULL)
	{
		MsgBox(g_szErrOom);
		goto Exit;
	}

	/* Read entire data section */
	if (!ReadFile(hFileInput, pDataFile, dwDataFileSize, &dwBytesRead, NULL) ||
		dwBytesRead != dwDataFileSize)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Initialize the LZO library */
	if (lzo_init() != LZO_E_OK)
	{
		MsgBox(g_szErrDecomp);
		goto Exit;
	}

	/* Decompress the data section */
	dwBytesRead = dwDataBlockSize;
	if (lzo1x_decompress_safe(pDataFile, dwDataFileSize, pDataBlock,
		&dwBytesRead, NULL) != LZO_E_OK || dwBytesRead != dwDataBlockSize)
	{
		MsgBox(g_szErrDecomp);
		goto Exit;
	}

	/* Quick and dirty search for lightmap chunk */
	bFound = FALSE;
	dwData = 0x0304303D;
	for (pData = pDataBlock; pData <= (pDataBlock + dwDataBlockSize - sizeof(dwData)); pData++)
	{
		if (!memcmp(pData, &dwData, sizeof(dwData)))
		{
			bFound = TRUE;
			break;
		}
	}

	if (!bFound)
	{
		MsgBox(g_szErrChunk);
		goto Exit;
	}

	pData += 4;

	/* Skip 'SKIP' */
	if (memcmp(pData, "PIKS", 4))
	{
		MsgBox(g_szErrChunk);
		goto Exit;
	}

	pData += 4;

	/* Get chunk size */
	memcpy(&dwFileSize, pData, sizeof(dwFileSize));
	if (dwFileSize == 0 || dwFileSize > dwDataBlockSize)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	pData += 4;

	/* Get boolean */
	memcpy(&dwData, pData, sizeof(dwData));
	if (dwData == 0)
	{
		MsgBox(g_szErrVersion);
		goto Exit;
	}
	
	pData += 4;

	/* Get version */
	memcpy(&dwVersion, pData, sizeof(dwVersion));
	if (dwVersion < 2 || dwVersion > 8)
	{
		MsgBox(g_szErrVersion);
		if (dwVersion < 2) goto Exit;
	}

	pData += 4;

	if (dwVersion >= 5)
	{
		/* Number of frames */
		memcpy(&dwFrames, pData, sizeof(dwFrames));
		if (dwFrames == 0)
		{
			MsgBox(g_szErrNone);
			goto Exit;
		}
		else if (dwFrames > 100)
		{
			MsgBox(g_szErrRead);
			goto Exit;
		}

		pData += 4;
	}

	while (dwFrames--)
	{
		if (!WriteLightMapFile(szFileName, &pData, &dwCount))
			goto Exit;

		if (dwVersion >= 3)
		{
			if (!WriteLightMapFile(szFileName, &pData, &dwCount))
				goto Exit;
		}

		if (dwVersion >= 6)
		{
			if (!WriteLightMapFile(szFileName, &pData, &dwCount))
				goto Exit;
		}
	}

	if (dwCount == 0)
		MsgBox(g_szErrNone);

Exit:
	/* Cleanup */
	if (pDataBlock != NULL)
		GlobalFreePtr(pDataBlock);
	if (pDataFile != NULL)
		GlobalFreePtr(pDataFile);
	if (hFileInput != NULL)
		CloseHandle(hFileInput);

	return 0;
}

/**********************************************************************************************/

BOOL WriteLightMapFile(PTSTR pszFileName, PBYTE* pData, DWORD* dwCount)
{
	TCHAR szFile[MAX_PATH];
	TCHAR szFormat[MAX_PATH];

	/* Read file size */
	DWORD dwFileSize = 0;
	memcpy(&dwFileSize, *pData, sizeof(dwFileSize));

	*pData += 4;

	if (dwFileSize == 0)
		return TRUE;

	/* Build filename */
	_tcsncpy(szFile, pszFileName, MAX_PATH - 1);

	_sntprintf(szFormat, _countof(szFormat), g_szFormat, ++(*dwCount));
	_tcsncat(szFile, szFormat, MAX_PATH);

	if (!memcmp(*pData, "RIFF", 4))
		_tcsncat(szFile, g_szWebp, MAX_PATH);
	else if ((*pData)[0] == 0xFF && (*pData)[1] == 0xD8)
		_tcsncat(szFile, g_szJpeg, MAX_PATH);
	else
		_tcsncat(szFile, g_szGbx, MAX_PATH);

	/* Create file */
	HANDLE hFileOutput = CreateFile(szFile, GENERIC_READ | GENERIC_WRITE,
		0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFileOutput == INVALID_HANDLE_VALUE)
	{
		MsgBox(g_szErrCreate);

		*pData += dwFileSize;

		return FALSE;
	}

	/* Write file */
	DWORD dwBytesWritten = 0;
	if (!WriteFile(hFileOutput, *pData, dwFileSize, &dwBytesWritten, NULL) ||
		dwBytesWritten != dwFileSize)
	{
		CloseHandle(hFileOutput);
		hFileOutput = NULL;
		DeleteFile(szFile);

		MsgBox(g_szErrWrite);

		*pData += dwFileSize;

		return FALSE;
	}

	CloseHandle(hFileOutput);
	hFileOutput = NULL;

	*pData += dwFileSize;

	return TRUE;
}

/**********************************************************************************************/
